/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const EXPORTED_SYMBOLS = ["NewWindow",
                          "getAboutUriChannel",
                          "util2",  // about:multifox
                          "Profile" // about:multifox
                         ];
const Cc = Components.classes;
const Ci = Components.interfaces;

const UnknownIdentity = 0;
const DefaultIdentity = 1;

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://multifox-modules/util.js");

util.log("===>LOADING main.js");


const NewWindow = {
  newId: function(win) {
    var id;
    if (this._shouldBeDefault(win)) {
      id = DefaultIdentity;
    } else {
      id = Profile.lowerAvailableId(win);
    }
    util.log("newIdentity " + id);
    Profile.save(win, id);
  },

  inheritId: function(newWin, prevWin) {
    util.log("inheritedIdentity");
    var id;
    if (this._shouldBeDefault(newWin)) {
      id = DefaultIdentity;
    } else {
      if (prevWin) {
        id = Profile.getIdentity(prevWin.document);
      } else {
        id = UnknownIdentity;
      }
    }
    Profile.save(newWin, id);
    util.log("/inheritedIdentity " + id);
  },

  applyRestore: function(win) {
    // restore: window is first configured by NewWindow.inheritId
    util.log("applyRestore");
    win.document.setUserData("multifox-dom-restore-fired", "ok", null);

    if (Profile.getIdentity(win.document) !== DefaultIdentity) {
      BrowserWindow.unregister(win);
      removeIcon(win.document);
    } else {
      // DefaultIdentity
      removeIcon(win.document);
    }

    var stringId = Cc["@mozilla.org/browser/sessionstore;1"]
                    .getService(Ci.nsISessionStore)
                    .getWindowValue(win, "multifox-dom-identity-id");
    var id = Profile.toInt(stringId);
    if (id < DefaultIdentity) { //UnknownIdentity?
      id = DefaultIdentity;
    }
    BrowserWindow.register(win, id);
  },

  _shouldBeDefault: function(win) {
    // popup opened by an extension (like GTB)
    //var chromeHidden = win.document.documentElement.getAttribute("chromehidden");
    //return chromeHidden.indexOf("location") > -1;
    return false;
  }

};


const Profile = {
  getIdentity: function(doc) {
    var icon = getIconNode(doc);
    if (icon === null) {
      var type = doc.documentElement.getAttribute("windowtype");
      if (type === "navigator:view-source") {
        var winOpener = doc.defaultView.opener;
        if (winOpener) {
          icon = getIconNode(winOpener.document);
        }
        util.log("view-source opener=" + winOpener + (icon ? icon.getUserData("multifox-dom-identity-id") : "[?]"));
      }
    }

    if (icon === null) {
      return DefaultIdentity;
    } else {
      var id = icon.getUserData("multifox-dom-identity-id");
      return this.toInt(id);
    }
  },

  save: function(win, id) {
    util.log("save " + id);
    if (Profile.getIdentity(win.document) !== DefaultIdentity) {
      BrowserWindow.unregister(win);
      removeIcon(win.document);
    }

    BrowserWindow.register(win, id);
    new SaveToSessionStore(win.document);
  },

  activeIdentities: function(ignoreWin) {
    var winEnum = util2.browserWindowsEnum();
    var arr = [];
    while (winEnum.hasMoreElements()) {
      var win = winEnum.getNext();
      if (ignoreWin !== win) {
        var id = Profile.getIdentity(win.document);
        if (arr.indexOf(id) === -1) {
          arr.push(id);
        }
      }
    }
    return arr;
  },

  lowerAvailableId: function(ignoreWin) {
    var arr = this.activeIdentities(ignoreWin); //ignore win -- it doesn't have a session id yet
    var id = DefaultIdentity;
    while (arr.indexOf(id) > -1) {
      id++;
    }
    return id; // id is available
  },

  toInt: function(str) {
    var rv = parseInt(str, 10);
    return isNaN(rv) ? DefaultIdentity : rv;
  },

  toString: function(id) {
    switch (id) {
      //case UnknownIdentity:
      //  return "\u221e"; // ∞
      default:
        return id.toString();
    }
  }

};


function SaveToSessionStore(doc) {
  this._doc = doc;
  this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
  this._timer.init(this, 1300, Ci.nsITimer.TYPE_ONE_SHOT);
}

SaveToSessionStore.prototype = {
  _doc: null,
  _timer: null,

  _restored: function() {
    return this._doc.getUserData("multifox-dom-restore-fired") === "ok";
  },

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),

  observe: function(aSubject, aTopic, aData) {
    util.log("SaveToSessionStore timer!");
    var doc = this._doc;
    if (doc.defaultView === null) {
      util.log("_syncToSessionStore window closed");
      return;
    }

    if (this._restored()) {
      // cancel, icon was updated by saveRestore
      util.log("_syncToSessionStore CANCELED icon=" +
               Profile.getIdentity(doc) +
               " sstore=" +
               Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore).getWindowValue(doc.defaultView, "multifox-dom-identity-id"));
      return;
    }

    // ok, restore nao ocorreu (pode nem ocorrer)
    var ss = Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore);
    var val = Profile.getIdentity(doc);

    try {
      // overwrite any previous value if called twice
      ss.setWindowValue(doc.defaultView, "multifox-dom-identity-id", val);
    } catch (ex) {
      // keep trying
      util2.logEx("SaveToSessionStore FAIL", val, doc, doc.defaultView, doc.defaultView.state, ex);
      this._timer.init(this, 700, Ci.nsITimer.TYPE_ONE_SHOT);
      return;
    }

    if (val <= DefaultIdentity) { // UnknownIdentity OR DefaultIdentity
      ss.deleteWindowValue(doc.defaultView, "multifox-dom-identity-id");
    }

    util.log("_syncToSessionStore OK=" + Profile.getIdentity(doc));
  }
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const BrowserWindow = {

  register: function(win, profileId) {
    var doc = win.document;

    if (Profile.getIdentity(doc) !== DefaultIdentity) {
      util2.throwStack.go("already registered win " + Profile.getIdentity(doc));
    }

    util.log("BrowserWindow.register " + profileId);
    if (profileId === DefaultIdentity) {
      return;
    }

    if (util.networkListeners.active === false) {
      // first multifox window!
      multifoxStartup();
    }

    createIcon(doc, profileId);

    var cook = Cookies.getListeners(profileId);
    XblContent.registerWindow(win);

    win.addEventListener("MultifoxContentEvent_CookieGetter_Request", cook.cookieGetter, false, true);
    win.addEventListener("MultifoxContentEvent_CookieSetter", cook.cookieSetter, false, true);
    win.addEventListener("MultifoxContentEvent_FeatureNotSupported", onNotSupported, false, true);

    // some MultifoxContentEvent_* listeners are not called when
    // there are "unload" listeners with useCapture=true. o_O
    // But they are called if event listener is an anonymous function.
    win.addEventListener("unload", onUnloadChromeWindow, false);
  },


  // should keep id for session restore?
  unregister: function(win) {
    var doc = win.document;

    var idw = Profile.getIdentity(doc);
    util.log("BrowserWindow.unregister " + idw);

    if (idw === DefaultIdentity) {
      // nothing to unregister
      return;
    }


    XblContent.unregisterWindow(win);
    var sessions = Profile.activeIdentities(win);
    var cook = Cookies.removeListeners(idw, sessions);

    win.removeEventListener("MultifoxContentEvent_CookieGetter_Request", cook.cookieGetter, false);
    win.removeEventListener("MultifoxContentEvent_CookieSetter", cook.cookieSetter, false);
    win.removeEventListener("MultifoxContentEvent_FeatureNotSupported", onNotSupported, false);

    var onlyDefault = (sessions.length === 1) && (sessions[0] === DefaultIdentity);
    if (onlyDefault) {
      // no more multifox windows
      multifoxShutdown();
    }
    win.removeEventListener("unload", onUnloadChromeWindow, false);
  }
};


function onUnloadChromeWindow(evt) {
  var win = evt.currentTarget;
  BrowserWindow.unregister(win);
}


function multifoxStartup() {
  util.log("multifoxStartup");
  Cookies.start();
  util.networkListeners.enable(httpListeners.request, httpListeners.response);
  XblContent.init();
  WatchObjectsStatus.start();
  //toggleDefaultWindowUI(true);
  util.log("/multifoxStartup");
}

function multifoxShutdown() {
  util.log("multifoxShutdown");
  util.networkListeners.disable();
  XblContent.shutdown();
  Cookies.stop();
  WatchObjectsStatus.stop();
  //toggleDefaultWindowUI(false);
}

function onNotSupported(evt) {
  evt.stopPropagation();
  switch (evt.data) {
    case "__defineSetter__":
      var doc = evt.target;
      util2.logEx("__defineSetter__?",
                  doc.location,
                  "title[" + doc.title + "]",
                  doc.__defineSetter__,
                  doc.wrappedJSObject.__defineSetter__,
                  util2.findBrowser(doc.defaultView),
                  util2.findBrowser2(doc.defaultView)
      );
      return;
    case "document":
      var doc = evt.target;
      util2.logEx("document?",
                  doc,
                  doc ? doc.location : "?"
      );
      return;
  }

  Components.utils.import("resource://multifox-modules/error.js");
  showError(evt.target.defaultView, evt.data);
}


const WatchObjectsStatus = {
  _timer: null,

  start: function() {
    this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    this._schedule();
  },

  stop: function() {
    this._timer.cancel();
    this._timer = null;
  },

  _schedule: function() {
    this._timer.init(this, 90000, Ci.nsITimer.TYPE_ONE_SHOT);
  },

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),

  observe: function(aSubject, aTopic, aData) {
    var buf = "watchObjectsStatus\n";

    buf += "m_frameNodes.length="+ m_frameNodes.length + " jsd=" + (jsd._service ? jsd._service.isOn : "null");

    var qty = m_frameNodes.length;
    for (var idx = qty - 1; idx >= 0; idx--) {
      buf += "\n" + idx + " " + m_frameNodes[idx].src;
    }

    util.log(buf);
    this._schedule();
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function getIconNode(doc) {
  if (!doc) {
    util2.throwStack.go("getIconNode "+doc);
  }
  return doc.getElementById("multifox-icon");
}


function removeIcon(doc) {
  util.log("removeIcon");
  var icon = getIconNode(doc);
  if (icon) {
    icon.parentNode.removeChild(icon);
  }
}


function createIcon(doc, profileId) {
  util.log("createIcon " + profileId);
  if (getIconNode(doc)) {
    util2.throwStack.go("createIcon ERR new=" + profileId + " current=" + Profile.getIdentity(doc));
  }
  /*
  var ub = doc.getElementById("urlbar");//textbox
  var dropMarker = doc.getAnonymousElementByAttribute(ub, "anonid", "popupset");
  var pn = doc.getBindingParent(dropMarker);
  util.log("pn=" + pn.tagName);
  try {
    var icon2 = dropMarker.insertBefore(doc.createElement("box"), dropMarker.firstChild);
    //var spc = pn.appendChild(doc.createElement("spacer"));
    //spc.setAttribute("flex", "1");
    var icon2 = pn.appendChild(doc.createElement("box"));
    var lb2 = icon2.appendChild(doc.createElement("label"));
    lb2.setAttribute("value", profileId);
    var s = lb2.style;
    s.color = "red";
  } catch (ex) {
    util2.throwStack.go("ico " + ex);
  }
  */

  var bar = doc.getElementById("urlbar-icons");

  var icon = bar.appendChild(doc.createElement("box"));
  icon.setAttribute("hidden", "true");
  icon.setAttribute("id", "multifox-icon");
  icon.setUserData("multifox-dom-identity-id", profileId === DefaultIdentity ? null : profileId, null);

  if (profileId !== UnknownIdentity) {
    doc.defaultView.setTimeout(initIconCore, 400, icon, profileId);
  }
}


function initIconCore(icon, profileId) {
  var doc = icon.ownerDocument;
  if (icon !== getIconNode(doc)) {
    util.log("initIconCore != " + profileId + icon + getIconNode(doc));
    return;
  }

  icon.setAttribute("tooltiptext", "Multifox");
  icon.style.MozBorderRadius = "2px";
  icon.style.paddingLeft = "6px";
  icon.style.paddingRight = "6px";
  icon.style.background = "transparent url(chrome://multifox/content/icon.png) repeat-x";

  var lb = icon.appendChild(doc.createElement("label"));
  lb.setAttribute("value", Profile.toString(profileId));
  var s = lb.style;
  s.color = "white";
  s.fontWeight = "bold";
  s.fontStyle = "normal";
  s.textRendering = "optimizelegibility";
  s.textShadow = "1px 1px 1px black";

  initIconNormal(icon);
  icon.setAttribute("hidden", "false");
}


function openMultifoxPopup(evt) {
  var icon = evt.currentTarget; //or this
  var doc = icon.ownerDocument;
  var win = doc.defaultView;

  var profileId = Profile.getIdentity(doc);
  var p1 = util.getText("icon.panel.p1.label", Profile.toString(profileId));
  var p2 = util.getText("icon.panel.p2.label");


  var panel = doc.getElementById("multifox-popup");
  if (panel) {
    //bug
    win.alert("OPENMULTIFOXPOPUP ==>HIDING " + panel.state);
    panel.hidePopup();
    return;
  }

  panel = doc.getElementById("mainPopupSet").appendChild(doc.createElement("panel"));
  var hx = panel.appendChild(doc.createElement("hbox"));

  var img = hx.appendChild(doc.createElement("image"));
  var desc1 = panel.appendChild(doc.createElement("description"));
  var desc2 = panel.appendChild(doc.createElement("description"));

  img.setAttribute("src", "chrome://multifox/content/logo-popup.png");
  desc1.appendChild(doc.createTextNode(p1));
  desc2.appendChild(doc.createTextNode(p2));
  desc1.style.fontWeight = "bold";


  if (profileId === UnknownIdentity) {
    desc2.hidden = true;
  }


  // options
  /*
  var gbox = panel.appendChild(doc.createElement("groupbox"));
  var caption = gbox.appendChild(doc.createElement("caption"));
  var vboxOptions = gbox.appendChild(doc.createElement("vbox"));

  caption.setAttribute("label", "Opes do perfil de identidade");


  var hboxProfile = vboxOptions.appendChild(doc.createElement("hbox"));
  hboxProfile.setAttribute("align", "center");

  var labelProfile = hboxProfile.appendChild(doc.createElement("label"));
  labelProfile.setAttribute("control", "profileList");
  labelProfile.appendChild(doc.createTextNode("Perfil de identidade desta janela:"));


  var listProfile = hboxProfile.appendChild(doc.createElement("menulist"));
  listProfile.setAttribute("id", "profileList");

  var menuProfile = listProfile.appendChild(doc.createElement("menupopup"));
  for (var idx = 2; idx < 7; idx++) {
    var itemProfile = menuProfile.appendChild(doc.createElement("menuitem"));
    itemProfile.setAttribute("label", idx);
  }


  var cboxCookies = vboxOptions.appendChild(doc.createElement("checkbox"));
  cboxCookies.setAttribute("label", "Desativar cookies");

  var vboxButton = vboxOptions.appendChild(doc.createElement("hbox"));
  var butClear = vboxButton.appendChild(doc.createElement("button"));
  butClear.setAttribute("label", "Limpar dados do perfil");
  */



  // about button

  var h = panel.appendChild(doc.createElement("hbox"));
  var spc = h.appendChild(doc.createElement("spacer"));
  spc.flex = 1;
  var but = h.appendChild(doc.createElement("button"));


  but.setAttribute("label", util.getText("icon.panel.button.label", "Multifox"));
  but.addEventListener("command", function(evt) {
    var w = evt.target.ownerDocument.defaultView;
    w.openUILinkIn("about:multifox", "tab", false, null, null);
    panel.hidePopup();
  }, true);

  if (win.gBrowser.selectedBrowser.currentURI.spec === "about:multifox") {
    but.hidden = true;
  }

  win.setTimeout(function() {
    // workaround to make focus work
    but.focus();
  }, 70);

  // xxxTODO
  // if (aero) {
  //panel.style.MozAppearance = "tooltip";
  //panel.style.color = "InfoText";


  panel.setAttribute("id", "multifox-popup");
  panel.addEventListener("popuphidden", closeMultifoxPopup, true);
  panel.style.padding = "0.9em 1.3em 0.9em 1.3em";
  panel.style.width = "30em";
  panel.openPopup(icon, "after_end");

  initIconPressed(icon);
}



function closeMultifoxPopup(evt) {
  var doc = evt.target.ownerDocument;

  var panel = doc.getElementById("multifox-popup");
  panel.parentNode.removeChild(panel);

  initIconNormal(getIconNode(doc));
}


function onIconHover(evt) {
  var icon = getIconNode(evt.target.ownerDocument);
  switch (evt.type) {
    case "mouseover":
      icon.style.backgroundPosition = "0pt -32px";
      icon.querySelector("label").style.textShadow = "1px 1px 1px black, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
      break;
    case "mouseout":
      icon.style.backgroundPosition = "0pt 0pt";
      icon.querySelector("label").style.textShadow = "1px 1px 1px black";
      break;
  }
}


function initIconNormal(icon) {
  icon.ownerDocument.defaultView.setTimeout(function(){
    // workaround
    // otherwise, mousedown on icon will close and open popup, even with stopPropagation
    icon.addEventListener("mousedown", openMultifoxPopup, true);
  }, 0);
  icon.addEventListener("mouseover", onIconHover, true);
  icon.addEventListener("mouseout", onIconHover, true);
  icon.style.backgroundPosition = "0pt 0pt";
  icon.querySelector("label").style.textShadow = "1px 1px 1px black";
}

function initIconPressed(icon) { // openpopup
  icon.removeEventListener("mousedown", openMultifoxPopup, true);
  icon.removeEventListener("mouseover", onIconHover, true);
  icon.removeEventListener("mouseout", onIconHover, true);
  icon.style.backgroundPosition = "0pt -64px";
  icon.querySelector("label").style.textShadow = "1px 1px 1px black, 0 1px 5px white, 3px 0 5px white, 0 -1px 5px white, -3px 0 5px white";
}

/*
// show/hide [1] "icon"
function toggleDefaultWindowUI(show) {
  var winEnum = util2.browserWindowsEnum();
  while (winEnum.hasMoreElements()) {
    var doc = winEnum.getNext().document;
    if (Profile.getIdentity(doc) === DefaultIdentity) {
      if (show) {
        createIcon(doc, DefaultIdentity);
      } else {
        removeIcon(doc);
      }
    }
  }
}
*/
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// Most of this code can be removed as soon bug 342715 is fixed.

const XblContent = {

  init: function() {
    this._enableContentXbl(true);
    frameXblActivator.init();
  },

  shutdown: function() {
    frameXblActivator.shutdown();
    this._enableContentXbl(false);
  },

  _enableContentXbl: function(enable) {
    var uri = util2.stringToUri("chrome://multifox/content/bindings.css");
    var i = Ci.nsIStyleSheetService;
    var type = i.AGENT_SHEET;
    var sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(i);
    if (enable) {
      if (sss.sheetRegistered(uri, type) === false) {
        sss.loadAndRegisterSheet(uri, type);
      }
    } else {
      if (sss.sheetRegistered(uri, type)) {
        sss.unregisterSheet(uri, type);
      }
    }
  },

  registerWindow: function(win) {
    win.addEventListener("MultifoxContentEvent_ApplyXbl", this._applyXblListener, false, true);
    win.gBrowser.addTabsProgressListener(topXblActivator);
    win.gBrowser.addEventListener("pagehide", frameXblActivator.pageHideListener, false);
    win.gBrowser.addEventListener("DOMContentLoaded", frameXblActivator.domLoadedListener, false);//xxxdel?
  },

  unregisterWindow: function(win) {
    win.removeEventListener("MultifoxContentEvent_ApplyXbl", this._applyXblListener, false);
    win.gBrowser.removeTabsProgressListener(topXblActivator);
    win.gBrowser.removeEventListener("pagehide", frameXblActivator.pageHideListener, false);
    win.gBrowser.removeEventListener("DOMContentLoaded", frameXblActivator.domLoadedListener, false);
  },

  // element bind
  _applyXblListener: function(evt) {
    evt.stopPropagation();
    var contentDoc = evt.target;
    if (contentDoc.getUserData("multifox-doc-xbl") !== null) {
      util.log("DUP XBL " + contentDoc.location);
      return;
    }
    contentDoc.setUserData("multifox-doc-xbl", "ok", null);

    evt.preventDefault(); // flag as Multifox document
    var h = contentDoc.getElementsByTagName("head")[0];
    var w = contentDoc.documentElement.style.width;


    // remove document
    for (var idx = 0, len = m_frameNodes.length; idx < len; idx++) {
      if (m_frameNodes[idx].contentDocument === contentDoc) {
        util.log("_applyXblListener frame removed! " + m_frameNodes[idx].src + evt.target.location.href);
        removePendingFrame(idx);
        break;
      }
    }
  }
};


const topXblActivator = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIWebProgressListener, Ci.nsISupportsWeakReference]),
  onStateChange: function() {},
  onSecurityChange: function() {},
  onProgressChange: function() {},
  onStatusChange: function() {},
  onLinkIconAvailable: function() {},
  onLocationChange: function(aBrowser, aWebProgress, aRequest, aLocation) {
    var contentDoc = aWebProgress.DOMWindow.document;
    if (contentDoc.getUserData("multifox-doc-xbl") !== null) {
      return;
    }

    // for frames: it fires only if DOMWindow is different from top window
    if (!contentDoc.documentElement) {
      // html5.enable=true ==> FAIL
      var dummy = contentDoc.getElementsByTagName("head")[0];
      util2.logEx("onLocationChange", "documentElement=" + contentDoc.documentElement, contentDoc.location, dummy);
      return;
    }
    var dummy = contentDoc.documentElement.style.width; // {bug} xhtml?
  }
};


// frame/iframe starts loading
//   -current page unloads (about:blank for new elements)
//   -pagehide is triggered
//     -frame node added to m_frameNodes
//     -jsd on
//   -new page starts loading
//   -jsd.topLevelHook.onCall triggered
//     -loop m_frameNodes
//       -xbl binding (triggered by jsd)
//         -new document.cookie is defined
//         -frame removed from m_frameNodes
//         -jsd off
//   -script not found by jsd
//     -xbl binding (<body>)
//       -frame removed from m_frameNodes
//       -jsd off

var m_frameNodes = null;

function removePendingFrame(idx) {
  m_frameNodes.splice(idx, 1);
  if (m_frameNodes.length === 0) {
    jsd.disable(); // disable jsd asap
  }
}

const frameXblActivator = {
  init: function() {
    m_frameNodes = [];
    jsd.disable();
  },

  shutdown: function() {
    jsd.disable();
    m_frameNodes = null;
  },

  _scheduleGc: function() {
    if (this._gcTimer === null) {
      this._gcTimer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
      this._gcTimer.init(this._timerObserver, 200, Ci.nsITimer.TYPE_ONE_SHOT);
    }
  },

  _gcTimer: null,
  _timerObserver: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),
    observe: function(aSubject, aTopic, aData) {
      frameXblActivator.clearPendingFrames();
    }
  },

  clearPendingFrames: function() {
    var buf = "";
    this._gcTimer = null;

    for (var idx = m_frameNodes.length - 1; idx > -1; idx--) {
      buf += "\nidx=" + idx;

      var frm = m_frameNodes[idx];
      var winFrame = frm.contentWindow;

      if (!winFrame) { // null most of the time, sometimes it is undefined
        // win closed?
        removePendingFrame(idx);
        buf += "\nNULL/undefined " + frm.tagName + " src=" + frm.src + " innerHTML=" + frm.innerHTML;
        continue;
      }

      var browser = util2.findBrowser(winFrame);
      if (browser === null) {
        removePendingFrame(idx);
        buf += "\nDEL5 " + frm.tagName + " src=" + frm.src;
        continue;
      }

      var ifr = frm;
      do {
        var parentFrame = ifr.ownerDocument.defaultView.frameElement;
        var parentWin = parentFrame === null ? browser : parentFrame;
        var pos = parentWin.contentDocument.documentElement.compareDocumentPosition(ifr);
        if ((pos & 1) === 1) { // DOCUMENT_POSITION_DISCONNECTED=1
          removePendingFrame(idx);
          buf += "\nDEL2 " + pos + " " + ifr.tagName + " src=" + ifr.src + " " + ifr.clientTop;
        } else {
          buf += "\nWAIT " + pos + " " + ifr.tagName + " src=" + ifr.src + " " + ifr.clientTop;
        }
        ifr = parentFrame;
      } while (parentFrame !== null)
    }

    if (m_frameNodes.length > 0) {
      this._scheduleGc();
    }
  },

  activatePendingDocuments: function() {
    for (var idx = m_frameNodes.length - 1; idx >= 0; idx--) {
      var node = m_frameNodes[idx];
      if (node.contentDocument) {
        //var h = d.getElementsByTagName("head")[0];
        var w = node.contentDocument.documentElement.style.width;
      } else {
        removePendingFrame(idx);
        util.log("activatePendingDocuments DEL " + idx + " " + node.src);
      }
    }
  },

  domLoadedListener: {
    handleEvent: function(evt) {
      var doc = evt.target;
      var frm = doc.defaultView.frameElement;
      if (frm === null) {
        return;
      }

      var idx = m_frameNodes.indexOf(frm);
      if (idx > -1) {
        removePendingFrame(idx);
      }
    }
  },

  pageHideListener: {
    handleEvent: function(evt) {
      // this=tabbrowser
      var doc = evt.target;
      var frm = doc.defaultView.frameElement;
      if (frm === null) {
        return;
      }

      if (m_frameNodes.indexOf(frm) > -1) {
        util.log("PageHide: frame ja em m_frameNodes " + frm.src + m_frameNodes.indexOf(frm));
        return;
      }

      // frm.tagName="object","iframe"
      util.log("frame pagehide add " + frm.tagName  + " src=" + frm.src + " innerHTML=[" + frm.innerHTML + "]");
      m_frameNodes.push(frm);
      frameXblActivator._scheduleGc();
      jsd.enable();
    }
  }
};


const jsd = {
  _service: null,

  _hook: {
    QueryInterface: XPCOMUtils.generateQI([Ci.jsdICallHook]),
    onCall: function(frame, type) {
      if (type === Ci.jsdICallHook.TYPE_TOPLEVEL_START) {
        frameXblActivator.activatePendingDocuments();
      }
    }
  },

  enable: function() {
    if (this._service === null) {
      util.log("jsd ON");
      var i = Ci.jsdIDebuggerService;
      var service = Cc["@mozilla.org/js/jsd/debugger-service;1"].getService(i);
      service.topLevelHook = this._hook;
      service.on();
      service.flags |= i.DISABLE_OBJECT_TRACE;
      this._service = service;
    }
  },

  disable: function() {
    util.log("jsd OFF");
    if (this._service) {
      this._service.off();
    }
    this._service = null;
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const httpListeners = {
  request: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function(aSubject, aTopic, aData) {
      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);

      var domWindow = getChannelWindow(httpChannel);
      if (domWindow === null) {
        return;
      }

      //chromeWindow="navigator:browser","navigator:view-source"
      var chromeWindow = util2.findChromeWindow(domWindow);
      if (chromeWindow === null) {
        util.log("chromeWindow=null," + domWindow.document.location);
        return;
      }

      var profileId = Profile.getIdentity(chromeWindow.document);

      if (profileId === UnknownIdentity) {
        // popup via js/window.open
        if (domWindow.opener) {
          var chromeOpener = util2.findChromeWindow(domWindow.opener);
          profileId = chromeOpener ? Profile.getIdentity(chromeOpener.document) : -1;
          if (profileId > UnknownIdentity) {
            Profile.save(chromeWindow, profileId);
          }
        }
        util.log("request [" + profileId + "] id.opener="+ profileId);
      }

      if (profileId === DefaultIdentity) {
        // normal browsing window, other dialogs
        return;
      }

      var myHeaders = HttpHeaders.fromRequest(httpChannel);

      if (myHeaders["authorization"] !== null) {
        Components.utils.import("resource://multifox-modules/error.js");
        showError(domWindow, "HTTP Authorization");
        return;
      }

      if (profileId === UnknownIdentity) {
        httpChannel.setRequestHeader("Cookie", null, false);
        return;
      }

      var uri = toInternalUri(httpChannel.URI, profileId);
      var cook = Cookies.getCookie(false, uri);
      httpChannel.setRequestHeader("Cookie", cook, false);
    }
  },

  response: {
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),

    // nsIObserver
    observe: function(aSubject, aTopic, aData) {
      var httpChannel = aSubject.QueryInterface(Ci.nsIHttpChannel);
      var domWindow = getChannelWindow(httpChannel);
      if (domWindow === null) {
        return;
      }

      var chromeWindow = util2.findChromeWindow(domWindow);
      if (chromeWindow === null) {
        return;
      }

      var profileId = Profile.getIdentity(chromeWindow.document);
      if (profileId === DefaultIdentity) {
        return;
      }


      var myHeaders = HttpHeaders.fromResponse(httpChannel);
      if (myHeaders["www-authenticate"] !== null) {
        Components.utils.import("resource://multifox-modules/error.js");
        showError(domWindow, "HTTP Authorization");
        return;
      }

      var setCookies = myHeaders["set-cookie"];
      if (setCookies === null) {
        return;
      }

      // server sent "Set-Cookie"
      httpChannel.setResponseHeader("Set-Cookie", null, false);

      if (profileId === UnknownIdentity) {
        return;
      }


      var uri = toInternalUri(httpChannel.URI, profileId);
      var multiSetCookies = new SetCookieParser(setCookies, true);

      var len = len = multiSetCookies.length;
      var mySetCookies = new Array(len);

      for (var idx = 0; idx < len; idx++) {
        var aCookie = multiSetCookies.getCookieByIndex(idx);
        var realDomain = aCookie.getStringProperty("domain");
        if (realDomain.length > 0) {
          aCookie.defineMeta("domain", cookieInternalDomain(realDomain, profileId));
          mySetCookies[idx] = aCookie.toHeaderLine();
        } else {
          mySetCookies[idx] = aCookie.raw;
        }
      }

      var internalCookies = mySetCookies.join("\n");//multiSetCookies.toHeader();
      Cookies.setCookie(false, uri, internalCookies);
    }
  }
};


const HttpHeaders = {
  visitLoop: {
    values: null,
    visitHeader: function(name, value) {
      var n = name.toLowerCase();
      if (n in this.values) {
        this.values[n] = value;
      }
    }
  },

  fromRequest: function(request) {
    var nameValues = {
      //"cookie": null, //for debug only
      "authorization": null
    }
    this.visitLoop.values = nameValues;
    request.visitRequestHeaders(this.visitLoop);
    return nameValues;
  },

  fromResponse: function(response) {
    var nameValues = {
      "set-cookie": null,
      "www-authenticate": null
    }
    this.visitLoop.values = nameValues;
    response.visitResponseHeaders(this.visitLoop);
    return nameValues;
  }
};


function getChannelWindow(channel) {
  if (channel.notificationCallbacks) {
    try {
      return channel
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext)
              .associatedWindow;
    } catch (ex) {
      //util2.logEx("channel.notificationCallbacks ", channel.notificationCallbacks, channel.URI.spec, ex);
    }
  }

  if (channel.loadGroup && channel.loadGroup.notificationCallbacks) {
    try {
      return channel
              .loadGroup
              .notificationCallbacks
              .getInterface(Ci.nsILoadContext)
              .associatedWindow;
    } catch (ex) {
      util2.logEx("channel.loadGroup", channel.loadGroup, channel.URI.spec, ex);
    }
  }

  //var isChrome = context.associatedWindow instanceof Ci.nsIDOMChromeWindow;
  //return context.isContent ? context.associatedWindow : null;
  //util.log("LOAD CONTEXT FAIL " + channel.URI.spec);
  return null; // e.g. <link rel=prefetch> <link rel=next> ...
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const PREF_COOKIE_BEHAVIOR = "network.cookie.cookieBehavior";

var m_documentCookie = null;
const Cookies = {
  _service: null,
  _prefs: null,

  start: function() {
    m_documentCookie = {};
    this._service = Cc["@mozilla.org/cookieService;1"].getService().QueryInterface(Ci.nsICookieService);
    this._prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch2);
    this._prefListener.behavior = this._prefs.getIntPref(PREF_COOKIE_BEHAVIOR);
    this._prefs.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
  },

  stop: function() {
    m_documentCookie = null;
    this._service = null;
    this._prefs.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    this._prefs = null;
  },


  _prefListener: {
    behavior: -1,

    QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
    // aTopic=nsPref:changed aData=network.cookie.cookieBehavior
    observe: function(aSubject, aTopic, aData) {
      this.behavior = aSubject
                      .QueryInterface(Ci.nsIPrefBranch)
                      .getIntPref(PREF_COOKIE_BEHAVIOR);
      util.log("pref! " + aSubject + aTopic + aData + this.behavior);
    }

  },

  getListeners: function(sessionId) {
    var cook;
    if (("id" + sessionId) in m_documentCookie) {
      cook = m_documentCookie["id" + sessionId];
    } else {
      cook = new documentCookieListeners(sessionId);
      m_documentCookie["id" + sessionId] = cook;
    }
    return cook;
  },

  removeListeners: function(idw, sessions) {
    var cook = m_documentCookie["id" + idw];
    if (sessions.indexOf(idw) === -1) {
      delete m_documentCookie["id" + idw];
    }
    return cook;
  },

  setCookie: function(fromJs, uri, val) {
    if (this._prefListener.behavior === 0) {
      this._setCookie(fromJs, uri, val);
      return;
    }

    var p = this._prefs;
    p.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, 0);
    this._setCookie(fromJs, uri, val);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, this._prefListener.behavior);
    p.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
  },

  getCookie: function(fromJs, uri) {
    if (this._prefListener.behavior === 0) {
      return this._getCookie(fromJs, uri);
    }

    var p = this._prefs;
    p.removeObserver(PREF_COOKIE_BEHAVIOR, this._prefListener);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, 0);
    var cookie = this._getCookie(fromJs, uri);
    p.setIntPref(PREF_COOKIE_BEHAVIOR, this._prefListener.behavior);
    p.addObserver(PREF_COOKIE_BEHAVIOR, this._prefListener, false);
    return cookie;
  },

  _setCookie: function(fromJs, uri, val) {
    if (fromJs) {
      this._service.setCookieString(uri,
                                    null,
                                    val,
                                    null);
    } else {
      //setCookieString doesn't work for httponly cookies
      this._service.setCookieStringFromHttp(uri,   // aURI
                                            null,  // aFirstURI
                                            null,  // aPrompt
                                            val,   // aCookie
                                            null,  // aServerTime
                                            null); // aChannel
    }
  },

  _getCookie: function(fromJs, uri) {
    if (fromJs) {
      return this._service.getCookieString(uri,
                                           null);
    } else {
      return this._service.getCookieStringFromHttp(uri,   // aURI
                                                   null,  // aFirstURI
                                                   null); // aChannel
    }
  }
};



function documentCookieListeners(id) {
  this.cookieGetter = new documentCookieGetter(id);
  this.cookieSetter = new documentCookieSetter(id);
}

function documentCookieGetter(id) {
  this.id = id;
}

documentCookieGetter.prototype = {
  id: -1,
  handleEvent: function(evt) {
    evt.stopPropagation();

    var profileId = this.id;
    switch (profileId) {
      case UnknownIdentity:
        break;
      case DefaultIdentity:
        util2.throwStack.go("documentCookieSetter " + profileId);
        break;
      default:
        var contentDoc = evt.target;
        var contentUrl = contentDoc.location.href;
        var uri = toInternalUri(util2.stringToUri(contentUrl), this.id);

        var cookie2 = Cookies.getCookie(true, uri);
        var cookie = cookie2 === null ? "" : cookie2;

        var myEvent = contentDoc.createEvent("MessageEvent");
        myEvent.initMessageEvent( // send cookie value to content
          "MultifoxContentEvent_CookieGetter_Response",
          true,
          false,
          cookie,
          contentUrl,
          "",
          contentDoc.defaultView);
        var success = contentDoc.dispatchEvent(myEvent);
        return;
    }
  }
};

function documentCookieSetter(id) {
  this.id = id;
}

documentCookieSetter.prototype = {
  id: -1,
  handleEvent: function(evt) {
    evt.stopPropagation();
    var profileId = this.id;

    switch (profileId) {
      case UnknownIdentity:
        break;
      case DefaultIdentity:
        util2.throwStack.go("documentCookieSetter " + profileId);
        break;
      default:
        var contentDoc = evt.target;
        var contentUrl = contentDoc.location.href;
        var unsafeCookie = evt.data;
        var uri = toInternalUri(util2.stringToUri(contentUrl), profileId);
        Cookies.setCookie(true, uri, unsafeCookie);
        break;
    }
  }
};



function toInternalUri(uri, sessionId) {
  var u = uri.clone();
  if (sessionId > DefaultIdentity) {
    u.host = cookieInternalDomain(u.host, sessionId);
  } else {
    util2.throwStack.go("invalid profile " + sessionId);
  }
  return u;
}


function cookieInternalDomain(domain, id) {
  return domain + ".multifox-profile-" + id;
}



function toLines(txt) {
  return txt.split(/\r\n|\r|\n/);
}


function SetCookieParser(cookieHeader, isSetCookie) {
  this.m_hasMeta = isSetCookie;
  this._allCookies = [];
  if (cookieHeader !== null) {
    var lines = toLines(cookieHeader);
    var len = lines.length;
    if (isSetCookie) {
      for (var idx = 0; idx < len; idx++) {
        this.parseLineSetCookie(lines[idx]);
      }
    } else {
      for (var idx = 0; idx < len; idx++) {
        this.parseLineCookie(lines[idx]);
      }
    }
  }
}

SetCookieParser.prototype = {
  parseLineSetCookie: function(headerLine) {
    var unit = new CookieUnit(headerLine);
    var items = headerLine.split(";");

    for (var idx = 0, len = items.length; idx < len; idx++) {
      var pair = splitValueName(items[idx]);
      var name = pair[0];
      var value = pair[1]; // null ==> name=HttpOnly, secure etc

      if (idx === 0) {
        if (name.length > 0 && value !== null) {
          unit.defineValue(name, value);
        } else {
          util2.logEx("_allCookies invalid", name, value, headerLine);
          break;
        }
      } else {
        unit.defineMeta(name, value);
      }
    }

    this._allCookies.push(unit);
  },

  parseLineCookie: function(headerLine) {
    var items = headerLine.split(";");
    for (var idx = 0, len = items.length; idx < len; idx++) {
      var unit = CookieUnit(items[idx]);
      var pair = splitValueName(items[idx]);
      unit.defineValue(pair[0], pair[1]);
      this._allCookies.push(unit);
    }
  },

  /*
  toHeader: function() {
    var allCookies = this._allCookies;
    var len = allCookies.length;
    //var buf = [];
    var buf = new Array(len);
    for (var idx = 0; idx < len; idx++) {
      //if (allCookies[idx].value !== null) {
      buf[idx] = allCookies[idx].toHeaderLine();
      //}
    }
    return this.m_hasMeta ? buf.join("\n") : buf.join(";");
  },

  getCookie: function(name) {
    var aCookie = this._allCookies;
    for (var idx = 0, len = aCookie.length; idx < len; idx++) {
      if (name === aCookie[idx].name) {
        return aCookie[idx];
      }
    }
    return null;
  },

  forEach: function(fn) {
    var c = this._allCookies;
    for (var idx = 0, len = c.length; idx < len; idx++) {
      fn(c[idx]);
    }
  }
  */

  get length() {
    return this._allCookies.length;
  },

  getCookieByIndex: function(idx) {
    return this._allCookies[idx];
  }
};

function CookieUnit(line) {
  this._data = {//instance value
    "_src": line
  };
}

CookieUnit.prototype = {
  _data: null,

  clone: function() {
    var c = new CookieUnit();
    for (var n in this._data) {
      c._data[n] = this._data[n];
    }
    return c;
  },

  get name() {
    var rv = this._data["_name"];
    return rv ? rv : "";
  },

  get value() {
    var rv = this._data["_value"];
    return rv ? rv : "";
  },

  get raw() {
    return this._data["_src"];
  },

  defineValue: function(name, val) {
    this._data["_name"] = name;
    this._data["_value"] = val;
  },

  //"secure":
  //"httponly":
  hasBooleanProperty: function(name) {
    //name = name.toLowerCase();
    return name in this._data;
  },

  setBooleanProperty: function(name, def) {
    if (def) {
      this._data[name] = null;
    } else {
      delete this._data[name];
    }
  },

  //"expires":
  //"domain":
  //"path":
  getStringProperty: function(name) {
    var rv = this._data[name];
    return rv ? rv : "";
    //return rv || "";
  },

  defineMeta: function(name, val) {
    name = name.toLowerCase();
    switch (name) {
      case "expires":
      case "domain":
      case "path":
      case "secure":
      case "httponly":
        this._data[name] = val;
        break;
    }
  },

  toHeaderLine: function() {//toString()
    var buf = [this.name + "=" + this.value];
    var props;

    props = ["secure", "httponly"];
    for (var idx = 0, len = props.length; idx < len; idx++) {
      var propName = props[idx];
      if (this.hasBooleanProperty(propName)) {
        buf.push(propName.toUpperCase());
      }
    }

    props = ["expires", "path", "domain"];
    for (var idx = 0, len = props.length; idx < len; idx++) {
      var propName = props[idx];
      var val = this.getStringProperty(propName);
      if (val.length > 0) {
        buf.push(propName.toUpperCase() + "=" + val);
      }
    }

    return buf.join(";");
  }
};


function splitValueName(cookie) {
  var idx = cookie.indexOf("=");
  if (idx === -1) {
    return [cookie.trim(), null];
  }

  // "a=bcd=e".split("=",2) returns [a,bcd]
  //   "abcde".split("=",2) returns [abcde]


  // MY =
  // 012^-----idx=3 length=4

  // MY =a:1=6
  // 012^-----idx=3 length=9

  var pair = ["", ""];
  pair[0] = cookie.substring(0, idx).trim(); //trim(cookie.substring(0, idx));
  idx++;
  if (idx < cookie.length) {
    pair[1] = cookie.substring(idx);
  }

  return pair;
}
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const util2 = {

  findChromeWindow: function(domWindow) {
    if (!domWindow) {
      util2.logEx("findChromeWindow", domWindow);
    }

    var docShellTreeItem = domWindow
                            .QueryInterface(Ci.nsIInterfaceRequestor)
                            .getInterface(Ci.nsIWebNavigation)
                            .QueryInterface(Ci.nsIDocShellTreeItem);

    var win = docShellTreeItem
                .rootTreeItem
                .QueryInterface(Ci.nsIInterfaceRequestor)
                .getInterface(Ci.nsIDOMWindow);

    if (win instanceof Ci.nsIDOMChromeWindow) {
      // wrappedJSObject allows access to gBrowser etc
      // it looks like wrappedJSObject=undefined for chrome URLs
      return win.wrappedJSObject ? win.wrappedJSObject : win;
    }

    // it doesn't work sometimes (docShellTreeItem.treeOwner is null)
    // try a [probably] slower method
    var browser = this.findBrowser2(domWindow);
    return browser ? browser.ownerDocument.defaultView : null;
  },

  browserWindowsEnum: function() {
    return Cc["@mozilla.org/appshell/window-mediator;1"]
            .getService(Ci.nsIWindowMediator)
            .getEnumerator("navigator:browser");
  },

  findBrowser2: function(domWindow) {
    var topDoc = domWindow.top.document;
    var winEnum = util2.browserWindowsEnum();
    while (winEnum.hasMoreElements()) {
      var chrWin = winEnum.getNext();
      var idx = chrWin.gBrowser.getBrowserIndexForDocument(topDoc);
      if (idx > -1) {
        return chrWin.gBrowser.browsers[idx];
      }
    }
    return null;
  },

  findBrowser: function(domWindow) {
    var win = this.findChromeWindow(domWindow);
    if (win) {
      var idx = win.gBrowser.getBrowserIndexForDocument(domWindow.top.document);
      if (idx > -1) {
        return win.gBrowser.browsers[idx];
      }
    }
    util2.logEx("findBrowser=null, tab not found", domWindow.document.location, win, this.findBrowser2(domWindow));
    return null;
  },

  appName: function() {
    return Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo).name;
  },

  stringToUri: function(spec) {
    var io = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
    try {
      return io.newURI(spec, null, null);
    } catch (ex) {
      return null;
    }
  },

  logEx: function() {
    var buf = "logEx";// + this.logEx.caller + "\n";// + util2.throwStack._toString(false, Components.stack) + "===";
    for (var idx = 0, len = arguments.length; idx < len; idx++) {
      buf += "\n" + arguments[idx];// + "=" + this.caller[arguments[idx]];
    }

    buf += "\n====\n" + util2.throwStack._toString(false, Components.stack);
    util.log(buf);
  },

  throwStack: {

    go: function(txt) {
      util.log("Stack: " + txt + "\n" + this._toString(false, Components.stack));
      throw txt;
    },

    _toArray: function(webOnly, theStack) {
      var allItems = [];
      if (!theStack) {
        return allItems;
      }

      function StackItem() {}
      StackItem.prototype = {
        lang: "?",
        filename: "?",
        lineNumber: -1,
        name: "?"
      };

      // aparentemente  um bug que ocorre as vezes, todas as propriedades sao undefined
      if (theStack.languageName == undefined) {
        var item = new StackItem();
        item.name = theStack.toString();
        allItems.push(item);
      }

      // myStack.caller  quase sempre null, mas algumas vezes  undefined...
      for (var myStack = theStack; myStack; myStack = myStack.caller) {
        if (webOnly) {
          var n = myStack.filename;
          if (n === null) {
            continue;
          }

          if ((n.indexOf("http:") !== 0) && (n.indexOf("https:") !== 0)) {
            continue;
          }
        }


        var item = new StackItem();
        allItems.push(item);
        item.lang = myStack.languageName;
        item.filename = myStack.filename;
        item.lineNumber = myStack.lineNumber; // Valid line numbers begin at "1". "0" indicates unknown.
        item.name = myStack.name;
      }

      allItems.reverse();
      return allItems;
    },

    _toString: function(webOnly, theStack) {
      var arr = this._toArray(webOnly, theStack);
      var lines = new Array(arr.length);
      for (var idx = arr.length - 1, idx2 = 0; idx > -1; idx--) {
        var lang = webOnly ? "" : arr[idx].lang + " ";
        lines[idx2] = "[" + (idx + 1) + "] "
                    + lang
                    + arr[idx].filename + " "
                    + arr[idx].name + " ("
                    + arr[idx].lineNumber + ")";
        idx2++;
      }
      return lines.join("\n");
    }

  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Multifox.
 *
 * The Initial Developer of the Original Code is
 * Jeferson Hultmann <hultmann@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function getAboutUriChannel(uri) {
  var channel = Cc["@mozilla.org/network/io-service;1"]
                  .getService(Ci.nsIIOService)
                  .newChannel("chrome://multifox/content/about-multifox.html", null, null);
  channel.originalURI = uri;
  return channel;
}
